// Copyright (c) 2012, François Yvon, Anil Kumar Singh and Guillaume Wisniewski
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following
// conditions are met:
//
// 1. Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above
// copyright notice, this list of conditions and the following
// disclaimer in the documentation and/or other materials provided
// with the distribution.
//
// 3. Neither the name of the copyright holder nor the names of its
// contributors may be used to endorse or promote products derived from
// this software without specific prior written permission.

// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
// FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
// COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
// STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
// OF THE POSSIBILITY OF SUCH DAMAGE.

#include "ngrampost.h"

#include <sstream>

std::string escapeJsonString(const std::string& input) {
  std::ostringstream ss;
  for (std::string::const_iterator iter = input.begin(); iter != input.end(); iter++) {
    switch (*iter) {
    case '\\': ss << "\\\\"; break;
    case '"': ss << "\\\""; break;
    case '/': ss << "\\/"; break;
    case '\b': ss << "\\b"; break;
    case '\f': ss << "\\f"; break;
    case '\n': ss << "\\n"; break;
    case '\r': ss << "\\r"; break;
    case '\t': ss << "\\t"; break;
    default: ss << *iter; break;
    }
  }
  return ss.str();
}


// Normalize FST so that posteriors can be efficiently computed.
//
// Path normalization is described in Equation (5) of [de Gispert et al,
// 2013]. In addition, the lattice is also topologically sorted.
//
// Note that the projection onto the output should be done before
// composition.
//
// Parameters
// ----------
// - ngramLatticeFst: the input lattice
// - normalizedFst: the normalized version of ngramLatticeFst
bool normalizeFST(VectorFst<LogArc>& ngramLatticeFst, VectorFst<LogArc>& normalizedFst)
{

  //BOOST_LOG_TRIVIAL(trace) << "Removing epsilon transitions...";
  RmEpsilon(&ngramLatticeFst);

  //BOOST_LOG_TRIVIAL(trace) << "Determinizing..." << endl;
  Determinize<LogArc > (ngramLatticeFst, &normalizedFst);

  //BOOST_LOG_TRIVIAL(trace) << "Minimizing..." << endl;
  Minimize<LogArc >(&normalizedFst, NULL, kDelta);

  //BOOST_LOG_TRIVIAL(trace) << "Pushing weights..." << endl;
  Push<LogArc > (&normalizedFst, REWEIGHT_TO_INITIAL , kDelta, true);

  //BOOST_LOG_TRIVIAL(trace) << "Topologically sorting the FST..." << endl;
  return TopSort(&normalizedFst);
}

/// \function computePosteriors
/// Actual computation of posteriors in three steps
/// \args:
/// - the input normalized lattice
/// - the table to return:
/// - the per state alphas
/// - the per state and ngram alphas
/// - the actual posteriors
int computePosteriors(VectorFst<LogArc>& normalizedFst,
		      AMap& alphas,
		      NMap& ngramAlphas,
		      AMap& posteriors)
{

  LogArc::StateId numStates = normalizedFst.NumStates();
  const SymbolTable* inputSymbols = normalizedFst.InputSymbols();

  /// Initializations of all ngrams that exist in the graph
  ///
  for (StateIterator<VectorFst<LogArc> > siter(normalizedFst); !siter.Done(); siter.Next()) {
    LogArc::StateId stateId = siter.Value();
    for (ArcIterator<VectorFst<LogArc> > aiter(normalizedFst, stateId); !aiter.Done(); aiter.Next()) {
      const LogArc &arc = aiter.Value();
      posteriors[arc.ilabel] = LogArc::Weight::Zero();    /// Wait - should not this be insert() ?
    }
  }
  /// Initialize the alphas 0 everywhere except on first nodes
  for (int is = 0; is < numStates; is++) {
    alphas[is] = LogArc::Weight::Zero();
    if (is == normalizedFst.Start())
      alphas[is] = LogArc::Weight::One();                 /// Wait - should not this be insert () ?
  }

  for (StateIterator<VectorFst<LogArc> > siter(normalizedFst); !siter.Done(); siter.Next()) {

    LogArc::StateId stateId = siter.Value();

    for (ArcIterator<VectorFst<LogArc> > aiter(normalizedFst, stateId); !aiter.Done(); aiter.Next()) {
      const LogArc &arc = aiter.Value();
      LogWeight w = arc.weight;
      LogArc::StateId nextStateId = arc.nextstate;
      /// Line 3 of de Gispert and al. push the per state alpha - easy
      alphas[nextStateId] = Plus(Times(alphas[stateId], arc.weight), alphas[nextStateId]);

      /// This arc label has never reached nextStateId -- insert it with a null weight
      if (ngramAlphas[nextStateId].find(arc.ilabel) == ngramAlphas[nextStateId].end())
	ngramAlphas[nextStateId][arc.ilabel] = LogArc::Weight::Zero();
      /// Push perstate, per ngram alphas, this is line 6 of de Gispert and al
      ngramAlphas[nextStateId][arc.ilabel] = Plus(Times(alphas[stateId], arc.weight), ngramAlphas[nextStateId][arc.ilabel]);

      /// Now simply propagate the other ngrams -- loop lines 7 to 10
      for (AMapItr itr = ngramAlphas[stateId].begin(); itr != ngramAlphas[stateId].end(); itr++) {
	int arcILabel = itr->first;
	/// skip this arc's label, it has already be taken care of
	if(arcILabel == arc.ilabel) continue;

	if (ngramAlphas[nextStateId].find(arcILabel) == ngramAlphas[nextStateId].end())
	  ngramAlphas[nextStateId][arcILabel] = LogArc::Weight::Zero();

	/// this is line 10
	ngramAlphas[nextStateId][arcILabel] = Plus(ngramAlphas[nextStateId][arcILabel],
						   Times(ngramAlphas[stateId][arcILabel],arc.weight));
      }
    }

    /// Updating posteriors upon reaching a final states
    if (normalizedFst.NumArcs(stateId) == 0 || normalizedFst.Final(stateId) != LogArc::Weight::Zero()) {

      for (AMapItr itr = ngramAlphas[stateId].begin(); itr != ngramAlphas[stateId].end(); itr++) {
	int arcILabel = itr->first;
	/// This is line 13
	posteriors[arcILabel] = Plus(posteriors[arcILabel], Times(ngramAlphas[stateId][arcILabel], normalizedFst.Final(stateId)));
      }
    }
  }
  //BOOST_LOG_TRIVIAL(trace) << "Finished calculating posteriors...";

  return 0;
}

void savePosteriors(AMap& posteriors, VectorFst<LogArc>& normalizedFst, string& output_file) {
  const SymbolTable* inputSymbolsFinal = normalizedFst.InputSymbols();

  ofstream output(output_file.c_str());
  bool first = true;

  output << "{";
  for (AMapItr itr = posteriors.begin(); itr != posteriors.end(); itr++) {
    int arcILabel = itr->first;
    if (!first) {
      output << "," << endl;
    }
    output << "\"" << escapeJsonString(inputSymbolsFinal->Find(arcILabel)) << "\": " << exp(-posteriors[arcILabel].Value());
    first = false;}

  output << "}";
}
