// Copyright (c) 2012, François Yvon, Anil Kumar Singh and Guillaume Wisniewski
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following
// conditions are met:
//
// 1. Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above
// copyright notice, this list of conditions and the following
// disclaimer in the documentation and/or other materials provided
// with the distribution.
//
// 3. Neither the name of the copyright holder nor the names of its
// contributors may be used to endorse or promote products derived from
// this software without specific prior written permission.

// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
// FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
// COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
// STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
// OF THE POSSIBILITY OF SUCH DAMAGE.

#include <boost/program_options.hpp>
#include <boost/log/trivial.hpp>

#include "ngrampost.h"

#define BOOST_ALL_DYN_LINK

namespace po = boost::program_options;


int main(int argc, char** argv) {

  // parameters
  int n_paths;
  string filename;
  string output = "posteriors.json";

  po::options_description desc("Mandatory options");
  desc.add_options()
    ("help,h", "produce help message")
    ("npaths,n", po::value<int>(&n_paths)->default_value(-1), "number of best path to use to estimate the posteriors")
    ("input,i", po::value<string>(&filename)->required(), "name of the file containing the lattices")
    ("output,o", po::value<string>(&output), "name of the file in which posteriors will be stored")
    ;

  po::variables_map args;
  po::store(po::parse_command_line(argc, argv, desc), args);

  if (args.count("help")) {
    cout << desc << "\n";
    return 1;
  }

  try {
    po::notify(args);
  } catch(boost::program_options::required_option& e) {
    BOOST_LOG_TRIVIAL(fatal) << "missing mandatory option";
    abort();
  }
    
  // Load FST from file and prune it if necessary
  VectorFst<LogArc>* logFst;
  StdVectorFst fst;
  StdVectorFst nbest_fst;

  if ((logFst = VectorFst<LogArc>::Read(filename)) == NULL) {
    BOOST_LOG_TRIVIAL(fatal) << "could not load fst";
    abort();
  }

  if (logFst->InputSymbols() == NULL) {
    cerr << "no input symbols. The FST must be compiled with the --keep-isymbols option" << endl;
    abort();
  }
  
  if (n_paths !=  -1) {
    BOOST_LOG_TRIVIAL(info) << "pruning";
    Map(*logFst, &fst, new LogToStdMapper());
    BOOST_LOG_TRIVIAL(trace) << "computing shortest path";
    ShortestPath(fst, &nbest_fst, n_paths);
    Map(nbest_fst, logFst, new StdToLogMapper());
  }

  // normalize FST
  BOOST_LOG_TRIVIAL(info) << "normalizing FST";
  VectorFst<LogArc> normalizedFST;
  normalizeFST(*logFst, normalizedFST);

  // Compute posterior
  int order = 1;
  
  AMap alphas;
  NMap ngramAlphas;
  AMap posteriors;
  computePosteriors(normalizedFST, alphas, ngramAlphas, posteriors);
  savePosteriors(posteriors, normalizedFST, output);
}
